@@ -77,7 +77,7 @@ class BIMNode (GeometryNode):
         
         if getMetaData:
             self.get_metadata_from_ifc_path()
-            if getResource or self._resource:
+            if getResource or self._resource is not None:
                 self.get_metadata_from_resource()
 
 #---------------------PROPERTIES----------------------------
