@@ -9,6 +9,7 @@ Be sure to check the properties defined in the above classes to initialise the N
 from ast import Raise
 from distutils import extension
 from pathlib import Path
+from typing import Tuple
 import xml.etree.ElementTree as ET
 import cv2
 import PIL
@@ -25,6 +26,7 @@ from scipy.spatial.transform import Rotation as R
 from geomapi.nodes import Node
 import geomapi.utils as ut
 import geomapi.utils.geometryutils as gt
+import geomapi.utils.imageutils as it
 
 
 class ImageNode(Node):
@@ -64,6 +66,8 @@ class ImageNode(Node):
         self.imageWidth = None # (int) number of pixels
         self.imageHeight = None  # (int) number of pixels
         self.focalLength35mm = None # (Float) focal length in mm     
+        self.keypoints = None # (array) the image keypoints
+        self.descriptors = None# (array) the image features
 
         super().__init__(   graph= graph,
                             graphPath= graphPath,
@@ -327,6 +331,21 @@ class ImageNode(Node):
             return None
         return self._orientedBoundingBox
 
+    def get_image_features(self, featureType = "Orb", max = 1000) -> Tuple[np.array, np.array]:
+        """Get the keypoints and the descriptors of this Nodes Image resource
+
+        Args:
+            featureType (str, optional): The featuretype to detect, use: orb, sift. Defaults to "Orb".
+            max (int, optional): The max features to detect. Defaults to 1000.
+
+        Returns:
+            Tuple[np.array, np.array]: The keypoints and the descriptors
+        """
+
+        if(self.keypoints is None or self.descriptors is None):
+            self.keypoints, self.descriptors = it.get_features(self.resource, featureType, max = max)
+        return self.keypoints, self.descriptors
+
     def get_metadata_from_resource(self) ->bool:
         """Returns the metadata from a resource. \n
 
@@ -472,7 +491,9 @@ class ImageNode(Node):
         else:
             cx=width/2-0.5
             cy=height/2-0.5
-        return o3d.camera.PinholeCameraIntrinsic(width,height,fx,fy,cx,cy)
+        pinholeCameraIntrinsic = o3d.camera.PinholeCameraIntrinsic(width,height,fx,fy,cx,cy)
+        self.intrinsic_matrix = pinholeCameraIntrinsic.intrinsic_matrix
+        return pinholeCameraIntrinsic
 
     def get_metadata_from_exif_data(self) -> bool:
         """Returns the metadata from a resource. \n
@@ -699,6 +720,8 @@ class ImageNode(Node):
         """
         if self._cartesianTransform is not None:
             pass
+        elif getattr(self,'cartesianTransform',None) is not None:
+            self._cartesianTransform = np.reshape(self.cartesianTransform, (4,4))
         elif getattr(self,'_cartesianBounds',None) is not None:
             self._cartesianTransform=gt.get_cartesian_transform(cartesianBounds=self._cartesianBounds)
         elif getattr(self,'_orientedBounds',None) is not None:
