@@ -60,10 +60,11 @@ class MeshNode (GeometryNode):
 
         #initialisation functionality
         if getResource:
-            self.get_resource()
+            self.get_resource() 
         
         if getMetaData:
-            self.get_metadata_from_resource()
+            if getResource or self._resource is not None:
+                self.get_metadata_from_resource()
 
 #---------------------METHODS----------------------------
 
@@ -91,12 +92,12 @@ class MeshNode (GeometryNode):
         Returns:
             o3d.geometry.TriangleMesh or None
         """
-        if self.resource is not None and len(self.resource.triangles)>=2:
+        if self._resource is not None and len(self._resource.triangles)>=2:
             pass
         elif self.get_path():
             resource =  o3d.io.read_triangle_mesh(self.path)
             if len(resource.triangles)>2:
-                self._resource  =resource
+                self._resource = resource
         return self._resource  
 
     def save_resource(self, directory:str=None,extension :str = '.ply') ->bool:
