@@ -6,6 +6,9 @@ import os
 import os.path
 from typing import List
 
+from sklearn.neighbors import NearestNeighbors # to compute nearest neighbors
+import laspy # this is to process las point clouds
+
 import geomapi.utils as ut
 import geomapi.utils.geometryutils as gt
 import matplotlib.pyplot as plt
@@ -15,6 +18,10 @@ import xlsxwriter
 from colour import Color
 from geomapi.nodes import *
 
+def compute_nearest_neighbor_in_bim(las:laspy.LasData,ref_clouds:List[] resolution:float = 0.1, )
+
+    compute_nearest_neighbors(reference_points:np.array, query_points:np.array,n:int=1)
+
 
 def get_boundingbox_of_list_of_geometries(geometries:List[o3d.geometry.PointCloud]) -> np.array:
     """Determines the global boundingbox of a group of Node containing geometries.
@@ -370,20 +377,21 @@ def color_point_cloud_by_LOA(pointcloud: o3d.geometry.PointCloud, identities, di
                     np.asarray(pointcloud.colors)[i] = [0,1,0]
         return pointcloud
 
-def color_point_cloud_by_distance(pointcloud: o3d.geometry.PointCloud, identities, distances, buckets: int = 5, dmax:float = 0.1, byElement: bool = False):
-    """Colors the resulting point cloud of the LOA analysis in a gradient by distance between the matched points from the reference and the source (very slow)
+def color_point_cloud_by_distance(pointcloud: o3d.geometry.PointCloud, identities:np.array, distances:np.array, buckets: int = 5, dmax:float = 0.1, byElement: bool = False)->o3d.geometry.PointCloud:
+    """Colorizes the resulting point cloud of the LOA analysis in a gradient by distance between the matched points from the reference and the source (very slow).\n
+
+    **NOTE**: use sklearn to make this faster.
 
     Args:
-        pointcloud (o3d.geometry.PointCloud): Point cloud from the LOA determination or pointcloud matching its the returned indeces
-        identities (nx1-array): Array containing the identity of the distance between two matched points
-        distances (nx1 array): Array containing the distances between two matched points
-        buckets (int, optional): Number of intervals to be colored in. Defaults to 5.
-        dmax (float, optional): Distances higher then this distance will be ignored. Defaults to 0.1.
-        byElement (bool, optional): If the LOA must be computed per element of for the enitre cloud. Defaults to False.
-        
+        1. pointcloud (o3d.geometry.PointCloud): Point cloud from the LOA determination or pointcloud matching its the returned indeces.\n
+        2. identities (nx1-array): Array containing the identity of the distance between two matched points.\n
+        3. distances (nx1 array): Array containing the distances between two matched points.\n
+        4. buckets (int, optional): Number of intervals to be colored in. Defaults to 5.\n
+        5. dmax (float, optional): Distances higher then this distance will be ignored. Defaults to 0.1m.\n
+        6. byElement (bool, optional): If the LOA must be computed per element of for the enitre cloud. Defaults to False.\n
 
     Returns:
-        o3d.geometry.PointCloud()
+        o3d.geometry.PointCloud
     """
     max = np.max(np.asarray(distances))
     if max > dmax:
@@ -439,16 +447,16 @@ def color_point_cloud_by_distance(pointcloud: o3d.geometry.PointCloud, identitie
             bucket +=1
         return pointcloud
 
-def csv_by_LOA(directory, LOAs, visibility=None):
-    """Function to report the LOA analysis in a csv file
+def csv_by_LOA(directory:str, LOAs, visibility=None):
+    """Function to report the LOA analysis in a csv file.\n
 
     Args:
-        directory (path): directory where the report must be saved
-        LOAs (_type_): results of the LOA computation
-        visibility (_type_, optional): array containing the per element visibility
+        1. directory (path): directory where the report must be saved.\n
+        2. LOAs (_type_): results of the LOA computation.\n
+        3. visibility (_type_, optional): array containing the per element visibility.\n
 
     Returns:
-        returns true when succeded
+        returns true when succeded.
     """
     if not os.path.exists(directory):
         os.mkdir(directory)
