@@ -89,21 +89,119 @@ def crop_mesh_by_convex_hull(source:trimesh.Trimesh, cutters: List[trimesh.Trime
                 submesh=combined
                 submeshes.append(submesh)
     return submeshes
+
+def mesh_sample_points_uniformly(mesh:o3d.geometry.TriangleMesh,resolution:float=0.1)->o3d.geometry.PointCloud:
+    """Sample the surface of a open3d mesh with a fixed resolution.\n
+
+    Args:
+        mesh (o3d.geometry.TriangleMesh)
+        resolution (float, optional): spatial resolution of the generated point cloud. Defaults to 0.1m.
+
+    Returns:
+        o3d.geometry.PointCloud
+    """
+    area=mesh.get_surface_area()
+    count=int(area/(resolution*resolution))
+    return mesh.sample_points_uniformly(number_of_points=count)
+
+def pcd_get_normals(pcd:o3d.geometry.PointCloud)->np.array:
+    """Compute open3d point cloud normals if not already present.\n
+
+    Args:
+        pcd (o3d.geometry.PointCloud)
+
+    Returns:
+        np.array:
+    """
+    pcd.estimate_normals() if not pcd.has_normals() else None
+    return np.asarray(pcd.normals)
+
+def get_points_and_normals(pcd,transform:np.array=None,getNormals=False)-> Tuple[np.array,np.array]:
+    """Extract points from different point cloud formats. Optionally extract or generate normals and apply a rigid body transformation.
+
+    Args:
+        1. pcd (_type_): point cloud \n
+        2. transform (np.array[4,4], optional): Rigid body transformation. Defaults to None.\n
+        3. getNormals (bool, optional): Defaults to False.\n
+
+    Raises:
+        ValueError: Only open3d, laspy and pandas data formats are currently supported.\n
+
+    Returns:
+        Tuple[np.array,np.array]: points, normals
+    """
+    if 'LasData' in str(type(pcd)):
+        points= transform_points(pcd.xyz,transform) if transform is not None else pcd.xyz
+        normals= las_get_normals(pcd,transform) if getNormals else None
+    elif 'PointCloud' in str(type(pcd)):
+        pcd.transform(transform) if transform is not None else None
+        points=np.asarray(pcd.points) 
+        normals=pcd_get_normals(pcd) if getNormals else None
+    elif 'DataFrame' in str(type(pcd)):
+        pass
+        # TODO: implement dataframe query_points and query points
+    else:
+        raise ValueError('type(pcd) == o3d.geometry.PointCloud, laspy point cloud or pandas dataframe')
+    return points,normals
+
+def compute_nearest_neighbor_with_normal_filtering(query_points:np.array,
+                                                    query_normals:np.array, 
+                                                    reference_points:np.array,
+                                                    reference_normals:np.array, 
+                                                    n:int=5,
+                                                    distanceThreshold=None)->Tuple[np.array,np.array]:
+    """Compute index and distance to nearest neighboring point in the reference dataset.\n
+    For the normal fultering, the n closest neighbors are considered of which the correspodence with the best matching normal is retained. \n
+
+    **NOTE**:  The index of outliers is set to -1 if distanceTreshold is not None.\n
     
-def compute_nearest_neighbors(reference_points:np.array, query_points:np.array,n:int=1)->Tuple[np.array,np.array]:
+    Args:
+        1. query_points (np.array[n,3]): points to evaluate.\n
+        2. query_normals (np.array[n,3]): normals to evaluate.\n
+        3. reference_points (np.array[n,3]): reference points.\n
+        4. reference_normals (np.array[n,3]): reference normals.\n
+        5. n (int, optional): number of neighbors.\n
+        5. distanceTreshold (_type_, optional): Distance threshold for the nearest neighbors.. Defaults to None.
+
+    Returns:
+        Tuple[np.array,np.array]: indices, distances 
+    """
+    #compute nearest neighbors 
+    indices,distances=compute_nearest_neighbors(query_points,reference_points,n=n)
+
+    #compute dotproduct
+    dotproducts=np.empty((indices.shape))
+    for i,ind in enumerate(np.hsplit(indices,indices.shape[1])):
+        dotproducts[:, i]= np.einsum('ij,ij->i', np.take(reference_normals, ind.flatten().T, axis=0), query_normals)
+    #select index with highest dotproduct
+    ind=np.argmax(np.absolute(dotproducts), axis=1)
+    indices = indices[np.arange(indices.shape[0]), ind]
+    distances = distances[np.arange(distances.shape[0]), ind]        
+    
+    #filter distances   
+    if distanceThreshold is not None:
+        assert distanceThreshold>0 ,f'distanceTreshold should be positive, got {distanceThreshold}'
+        indices=np.where(distances>distanceThreshold,-1,indices)
+        distances=np.where(distances>distanceThreshold,-1,distances)  
+        
+    return indices,distances
+
+def compute_nearest_neighbors(query_points:np.array,reference_points:np.array, n:int=1)->Tuple[np.array,np.array]:
     """Compute nearest neighbors (indices and distances) from querry to reference points.
 
     Args:
-        1. reference_points (np.array): points used as a reference.\n
-        2. query_points (np.arrayn): points to calculate the distance from.\n 
+        
+        1. query_points (np.arrayn): points to calculate the distance from.\n 
+        2. reference_points (np.array): points used as a reference.\n        
         3. n (int, optional): number of neighbors. Defaults to 1.\n
+        4. maxDist (float, optional): max distance to search. \n
 
     Returns:
-        Tuple[np.array,np.array]: distances, indices 
+        np.array[n,2]: indices, distances
     """    
     nbrs = NearestNeighbors(n_neighbors=n, algorithm='kd_tree').fit(reference_points)
-    distances, indices = nbrs.kneighbors(query_points)
-    return indices, distances
+    distances,indices, = nbrs.kneighbors(query_points)
+    return indices,distances
 
 def filter_pcd_by_distance(sourcePcd : o3d.geometry.PointCloud, testPcd: o3d.geometry.PointCloud, maxDistance : float) -> Tuple[o3d.geometry.PointCloud,o3d.geometry.PointCloud]:
     """Splits the sourcePcd in close and too far points compared to the testPcd based on the Euclidean distance.\n
@@ -205,7 +303,6 @@ def get_mesh_collisions_trimesh(sourceMesh: o3d.geometry.TriangleMesh, geometrie
         # add all geometries to the collision manager
         collisionManager=trimesh.collision.CollisionManager()
         for idx,geometry in enumerate(geometries):
-            test=type(geometry)
             if 'TriangleMesh' in str(type(geometry)) and len(geometry.triangles) >1:
                 referenceTrimesh=mesh_to_trimesh(geometry)
                 collisionManager.add_object(idx,referenceTrimesh)
@@ -456,7 +553,7 @@ def las_add_extra_dimensions(las:laspy.LasData, recordData:np.array, names:List[
     
     Args:
         1. las (laspy.LasData): las point cloud (laspy API).\n
-        2. recordData (np.array[len(pcd.points),n] or pandas.DataFrame): array or dataframe with len(columns) == len(pcd.points).\n
+        2. recordData (tuple with np.array[len(pcd.points)] or pandas.DataFrame): array or dataframe with len(columns) == len(pcd.points).\n
         3. names (str, optional): dimension names. Defaults to ['newField'].\n
         4. dtypes (str, optional): types of the columns. Defaults to ["uint8"].\n
 
@@ -466,26 +563,36 @@ def las_add_extra_dimensions(las:laspy.LasData, recordData:np.array, names:List[
     #validate inputs
     names=ut.item_to_list(names)
     dtypes=ut.item_to_list(dtypes)
-    assert 'las' in type(las)
+    assert 'las' in str(type(las))
     assert len(names)==len(dtypes)    
-    if 'np.array' in type(recordData):
-        assert recordData.shape[0]==las.header.point_count
-        assert recordData.shape[1]==len(names)
-    elif 'dataframe' in type(recordData):
-        assert len(recordData)==las.header.point_count
-        assert len(recordData.columns)==len(names)
-    
+    if 'array' in str(type(recordData)):
+        assert recordData.shape[0]==las.header.point_count ,f'one of the recordData contains entries != (len(las.xyz)).'
+        assert recordData.shape[1]==len(names), f'recordData.shape[1] !=len(names).'
+    elif 'dataframe' in str(type(recordData)):
+        assert len(recordData)==las.header.point_count,f'one of the recordData contains entries != (len(las.xyz)).'
+        assert len(recordData.columns)==len(names), f'recordData.shape[1] !=len(names).'
+    elif 'Tuple' in str(type(recordData)):
+        assert len(recordData)==len(names), f'len(recordData) !=len(names).'
+        
     #create dimensions
     extraBytesParams=[laspy.ExtraBytesParams(name=name, type=dtype) for name,dtype in zip(names,dtypes)]
-    las.add_extra_dims(extraBytesParams)
-    
+    las.add_extra_dims(extraBytesParams)   
+
     #add data
-    if 'np.array' in type(recordData):
-        [setattr(las,name,recordData[:,i]) for i,name in enumerate(names)]
-    elif 'dataframe' in type(recordData):
+    if 'array' in str(type(recordData[0])):
+        [setattr(las,name,recordData[i]) for i,name in enumerate(names)]
+    elif 'dataframe' in str(type(recordData)):
         [setattr(las,name,recordData[[i]]) for i,name in enumerate(names)]
     return las
 
+def las_get_normals(las:laspy.LasData,transform:np.array=None) -> np.array:
+    if all(n in ['normalX','normalY','normalZ'] for n in list(las.point_format.dimension_names)):
+        normals=np.hstack((las['normalX'],las['normalY'],las['normalZ']))
+        normals=transform_points(normals,transform) if transform is not None else normals
+    else:
+        normals=np.asarray(las_to_pcd(las,transform,getColors=False,getNormals=True).normals)    
+    return normals
+
 def las_to_pcd(las:laspy.LasData, transform:np.array=None, getColors:bool=True,getNormals:bool=False)->o3d.geometry.PointCloud:
     """Converts a laspy point cloud to an open3d point cloud.\n
 
@@ -524,6 +631,15 @@ def las_to_pcd(las:laspy.LasData, transform:np.array=None, getColors:bool=True,g
             green=green/255
             blue=blue/255
         pcd.colors=o3d.utility.Vector3dVector(np.vstack((red,green,blue)).transpose())
+        
+    #compute normals
+    if getNormals:
+        if all(n in ['normalX','normalY','normalZ'] for n in list(las.point_format.dimension_names)):
+            normals=np.hstack((las['normalX'],las['normalY'],las['normalZ']))
+            newNormals=transform_points(normals,transform) if transform is not None else normals
+            pcd.normals=o3d.utility.Vector3dVector(newNormals)
+        else:
+            pcd.estimate_normals()
     return pcd
 
 def dataframe_to_pcd(df:pd.DataFrame,pointFields:List[str]=['x', 'y', 'z'],transform:np.array=None)->o3d.geometry.PointCloud:
@@ -903,7 +1019,33 @@ def box_to_mesh(box:o3d.geometry) ->o3d.geometry.TriangleMesh:
                         [3,4,5],
                         [3,6,4],
                         [4,6,7]])) 
-    return mesh                   
+    return mesh 
+                  
+def ifc_get_materials(ifcElements:List[ifcopenshell.entity_instance])-> List[str]: 
+    """Get ifc materials from an ifcElement
+
+    Args:
+        ifcElements (List[ifcopenshell.entity_instance])
+
+    Returns:
+        List[str]: names of materials
+    """
+    material_list=[]
+    for ifcElement in ut.item_to_list(ifcElements):
+        if ifcElement.HasAssociations:
+            for i in ifcElement.HasAssociations:
+                if i.is_a('IfcRelAssociatesMaterial'):
+                    if i.RelatingMaterial.is_a('IfcMaterial'):
+                        material_list.append(i.RelatingMaterial.Name)
+
+                    if i.RelatingMaterial.is_a('IfcMaterialList'):
+                        for materials in i.RelatingMaterial.Materials:
+                            material_list.append(materials.Name)
+
+                    if i.RelatingMaterial.is_a('IfcMaterialLayerSetUsage'):
+                        for materials in i.RelatingMaterial.ForLayerSet.MaterialLayers:
+                            material_list.append(materials.Material.Name)
+    return material_list
 
 def ifc_to_mesh(ifcElement:ifcopenshell.entity_instance)-> o3d.geometry.TriangleMesh: 
     """Convert an ifcOpenShell geometry to an Open3D TriangleMesh.\n
@@ -1414,9 +1556,11 @@ def crop_point_cloud_from_meshes(pcd: o3d.geometry.PointCloud,meshes:List[o3d.ge
             pcdRemaining=pcd.select_by_index(nonIndices) 
     return croppedPcds, indices
 
-def create_identity_point_cloud(geometries: o3d.geometry.PointCloud, resolution:float = 0.1) -> Tuple[o3d.geometry.PointCloud, np.array]:
+def create_identity_point_cloud(geometries: o3d.geometry.PointCloud, resolution:float = 0.1, getNormals=False) -> Tuple[o3d.geometry.PointCloud, np.array]:
     """Returns a sampled point cloud colorized per object of a set of objects along with an array with an identifier for each point.
 
+    TODO: MB also store normals 
+    
     Args:
         1.geometries (o3d.geometry.PointCloud or o3d.geometry.TriangleMesh) \n
         2.resolution (float, optional): (down)sampling resolution for the point cloud. Defaults to 0.1.\n
@@ -1435,6 +1579,7 @@ def create_identity_point_cloud(geometries: o3d.geometry.PointCloud, resolution:
     for i,geometry in enumerate(geometries):
         if 'PointCloud' in str(type(geometry)) :
             pcd=geometry.voxel_down_sample(resolution)
+            get_points_and_normals(pcd,getNormals=getNormals) if getNormals else None
             indentityArray=np.vstack((indentityArray,np.full((len(pcd.points), 1), i)))
             pcd.paint_uniform_color(colorArray[i])
             identityPointCloud +=pcd
@@ -1442,8 +1587,8 @@ def create_identity_point_cloud(geometries: o3d.geometry.PointCloud, resolution:
         elif 'TriangleMesh' in str(type(geometry)):
             area=geometry.get_surface_area()
             count=int(area/(resolution*resolution))
-            pcd=geometry.sample_points_uniformly(number_of_points=10*count)
-            pcd=pcd.voxel_down_sample(resolution)
+            pcd=geometry.sample_points_uniformly(number_of_points=count, use_triangle_normal=getNormals)
+            # pcd=pcd.voxel_down_sample(resolution) #! why 10*count and downsample?
             indentityArray=np.vstack((indentityArray,np.full((len(pcd.points), 1), i)))
             pcd.paint_uniform_color(colorArray[i])
             identityPointCloud +=pcd
@@ -2037,7 +2182,22 @@ def save_view_point(geometry: o3d.geometry, filename:str) -> None:
     o3d.io.write_pinhole_camera_parameters(filename, param)
     vis.destroy_window()
 
-def execute_global_registration(source_down, target_down, source_fpfh,target_fpfh, voxel_size):
+# OPEN3D global registration
+
+def execute_global_registration(source_down : o3d.geometry.Geometry, target_down : o3d.geometry.Geometry, source_fpfh : np.array,target_fpfh : np.array, voxel_size : float):
+    """Performs a RANSAC based Fpfh feature matching of 2 pointclouds
+
+    Args:
+        source_down (o3d.geometry.Geometry): the source PCD
+        target_down (o3d.geometry.Geometry): the Target PCD
+        source_fpfh (np.array): The sourse features
+        target_fpfh (np.array): The target features
+        voxel_size (np.array): The size to downsample to for the feature matching
+
+    Returns:
+        RegistrationResult: The resulting registration
+    """
+
     distance_threshold = voxel_size * 1.5
     print(":: RANSAC registration on downsampled point clouds.")
     print("   Since the downsampling voxel size is %.3f," % voxel_size)
@@ -2051,7 +2211,20 @@ def execute_global_registration(source_down, target_down, source_fpfh,target_fpf
 
     return result
 
-def execute_fast_global_registration(source_pcd, target_pcd, source_fpfh,target_fpfh, radius):
+def execute_fast_global_registration(source_pcd : o3d.geometry.Geometry, target_pcd : o3d.geometry.Geometry, source_fpfh : np.array,target_fpfh : np.array, radius: float):
+    """Performs a RANSAC based Fpfh feature matching of 2 pointclouds using a faster algorithm
+
+    Args:
+        source_down (o3d.geometry.Geometry): the source PCD
+        target_down (o3d.geometry.Geometry): the Target PCD
+        source_fpfh (np.array): The sourse features
+        target_fpfh (np.array): The target features
+        radius (float): The correspondance radius
+
+    Returns:
+         RegistrationResult: The resulting registration
+    """
+
     print(":: Apply fast global registration with distance threshold %.3f" \
             % radius)
     result = o3d.pipelines.registration.registration_fgr_based_on_feature_matching(
@@ -2059,4 +2232,36 @@ def execute_fast_global_registration(source_pcd, target_pcd, source_fpfh,target_
         o3d.pipelines.registration.FastGlobalRegistrationOption(
             maximum_correspondence_distance=radius))
 
-    return result
\ No newline at end of file
+    return result
+
+# Average multiple quaternions with specific weights
+# The weight vector w must be of the same length as the number of rows in the
+# quaternion maxtrix Q
+#source: https://github.com/christophhagen/averaging-quaternions/
+def weighted_average_quaternions(Q : np.array, w: np.array):
+    """Calculates a weighted average of a list of quaternions 
+    The weight vector w must be of the same length as the number of rows in the quaternion maxtrix Q
+
+    Args:
+        Q (List[np.array]): a list of quaternions
+        w (List[float]): a lost of weights
+
+    Returns:
+        np.array: the weighted 4x1 quaternion
+    """
+    # Number of quaternions to average
+    M = Q.shape[0]
+    A = np.zeros(shape=(4,4))
+    weightSum = 0
+    for i in range(0,M):
+        q = Q[i,:]
+        A = w[i] * np.outer(q,q) + A
+        weightSum += w[i]
+    # scale
+    A = (1.0/weightSum) * A
+    # compute eigenvalues and -vectors
+    eigenValues, eigenVectors = np.linalg.eig(A)
+    # Sort by largest eigenvalue
+    eigenVectors = eigenVectors[:,eigenValues.argsort()[::-1]]
+    # return the real part of the largest eigenvector (has only real part)
+    return np.real(eigenVectors[:,0].A1)
\ No newline at end of file
