@@ -1,5 +1,5 @@
 """
-Image Utils - functions for handling 360° panoramic images.
+Image Utils - functions for handling images.
 
 Orientation is given as a quaternion in scalar-last format [x, y, z, w]. For an
 equirectangular pano format this is the orientation of the center of the pano.
@@ -8,6 +8,7 @@ The coordinate system is right-handed with z as the vertical direction.
 
 
 from datetime import datetime
+from typing import Tuple
 import xml.etree.ElementTree as ET
 from typing import List, Tuple
 
@@ -108,7 +109,171 @@ def calibrate_camera(images, CheckerBoardPattern = (7,7), squareSize: float = 1,
 
     return ret, mtx, dist, rvecs, tvecs
     
+def get_features(img : np.array, featureType : str = "Orb", max : int = 1000):
+    """Compute the image features and descriptors
 
+    Args:
+        img (np.array): The source image
+        method (str, optional): The type of features to detect, choose between: orb, sift or fast. Defaults to "Orb".
+        max (int, optional): The destection treshold. Defaults to 1000.
+
+    Raises:
+        ValueError: If the provided method is incorrect
+
+    Returns:
+        Keypoints, descriptors: The detected keypoints and descritors of the source image
+    """
+
+    im1Gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
+
+    if(featureType.lower() == "orb"):
+        detector = cv2.ORB_create(max)
+    if(featureType.lower() == "sift"):
+        detector = cv2.SIFT_create(max)
+    if(featureType.lower() == "fast"):
+        detector = cv2.FastFeatureDetector_create(max)
+    else:
+        raise ValueError("Invalid method name, please use one of the following: orb, sift, fast")
+    
+    return detector.detectAndCompute(im1Gray, None)
+
+def match_features(des1: np.array, des2: np.array, matchMethod: str = "bfm_orb", 
+                    k: int = 2, goodPercent: float = 0.75, checks: int = 100, 
+                    table_number: int = 12, key_size: int = 20, multi_probe_level:  int = 2):
+    """Matches 2 sets of features using different methods
+
+    Args:
+        des1 (np.array): The descriptors of the first image
+        des2 (np.array): The descriptors of the first image
+        method (str, optional): Use one of the following: bfm_orb, bfm_sift, flann_orb, flann_sift. Defaults to "bfm".
+        k (int, optional): The nr of best matches per descriptor. Defaults to 2.
+        goodPercent (float, optional): The filter value for the ratio test. Defaults to 0.75.
+        checks (int, optional): the number of times the trees in the index should be recursively traversed. Higher values gives better precision, but also takes more time. Defaults to 100.
+        table_number (int, optional): Flann orb parameter. Defaults to 12.
+        key_size (int, optional): Flann orb parameter. Defaults to 20.
+        multi_probe_level (int, optional): Flann orb parameter. Defaults to 2.
+
+    Raises:
+        ValueError: If the provided method is incorrect
+
+    Returns:
+        np.array: The matches
+    """
+    
+    if(matchMethod.lower() == "bfm_orb"):
+        # create BFMatcher object
+        bf = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)
+        # Match descriptors.
+        matches = bf.match(des1,des2)
+        # Sort them in the order of their distance.
+        matches = sorted(matches, key = lambda x:x.distance)
+    
+    if(matchMethod.lower() == "bfm_sift"):
+        # BFMatcher with default params
+        bf = cv2.BFMatcher()
+        matches = bf.knnMatch(des1,des2,k=k)
+        # Apply ratio test
+        good = []
+        for m,n in matches:
+            if m.distance < goodPercent*n.distance:
+                good.append([m])
+        matches = good
+    
+    if(matchMethod.lower() == "flann_sift"):
+        # FLANN parameters
+        FLANN_INDEX_KDTREE = 1
+        index_params = dict(algorithm = FLANN_INDEX_KDTREE, trees = 5)
+        search_params = dict(checks=checks)   # or pass empty dictionary
+        flann = cv2.FlannBasedMatcher(index_params,search_params)
+        matches = flann.knnMatch(des1,des2,k=k)
+        # Apply ratio test
+        good = []
+        for m,n in matches:
+            if m.distance < goodPercent*n.distance:
+                good.append([m])
+        matches = good
+    
+    if(matchMethod.lower() == "flann_orb"):
+        # FLANN parameters
+        FLANN_INDEX_LSH = 6
+        index_params= dict(algorithm = FLANN_INDEX_LSH,
+                        table_number = table_number, # 12
+                        key_size = key_size,     # 20
+                        multi_probe_level = multi_probe_level) #2
+        search_params = dict(checks=checks)   # or pass empty dictionary
+        flann = cv2.FlannBasedMatcher(index_params,search_params)
+        matches = flann.knnMatch(des1,des2,k=k)
+        # Apply ratio test
+        good = []
+        for m,n in matches:
+            if m.distance < goodPercent*n.distance:
+                good.append([m])
+        matches = good
+    
+    else:
+        raise ValueError("Invalid method name, please use one of the following: bfm_orb, bfm_sift, flann_orb, flann_sift")
+
+    return matches
+
+def match_images(img1: np.array, img2: np.array, featureType: str = "orb", matchMethod: str = "bfm"):
+    """Matches 2 images using different methods
+
+    Args:
+        img1 (np.array): The first image
+        img2 (np.array): The second image
+        featureType (str, optional): The type of feature to detect, use: orb, sift. Defaults to "orb".
+        matchMethod (str, optional): The matching method to use, use: bfm, flann. Defaults to "bfm".
+
+    Raises:
+        ValueError: if an incorrect feature or method name is provided
+
+    Returns:
+        np.array: The matches of the 2 images
+    """
+    
+    if(featureType.lower() == "orb"):
+        kp1, des1 = get_features(img1, "orb")
+        kp2, des2 = get_features(img2, "orb")
+        if(matchMethod.lower() == "bfm"):
+            return match_features(des1, des2, "bfm_orb")
+        if(matchMethod.lower() == "flann"):
+            return match_features(des1, des2, "flann_orb")
+        raise ValueError("Invalid method name, please use one of the following: bfm, flann")
+    
+    if(featureType.lower() == "sift"):
+        kp1, des1 = get_features(img1, "sift")
+        kp2, des2 = get_features(img2, "sift")
+        if(matchMethod.lower() == "bfm"):
+            return match_features(des1, des2, "bfm_sift")
+        if(matchMethod.lower() == "flann"):
+            return match_features(des1, des2, "flann_sift")
+        raise ValueError("Invalid method name, please use one of the following: bfm, flann")
+    
+    raise ValueError("Invalid feature name, please use one of the following: orb, sift")
+
+def create_transformation_matrix(R: np.array, t: np.array) -> np.array:
+    """Combines a rotation matrix and translation vector into a 4x4 transformation matrix
+
+    Args:
+        R (np.array): 3x3 rotation matrix
+        t (np.array): 3x1 translation vector
+
+    Returns:
+        np.array: 4x4 transformation matrix
+    """
+
+    return np.vstack((np.hstack((R,np.reshape(t,[3,1]))), np.array([0,0,0,1])))
+
+def split_transformation_matrix(T:np.array):
+    """Splits a transformation matrix into a rotation and translationmatrix
+
+    Args:
+        T (np.array): The 4x4 transormationmatrix
+
+    Returns:
+        np.array, np.array: 3x3 rotationmatrix R, 3x1 translationmatrix t
+    """
+    return T[:3,:3], T[:3,3]
 
 class AlignmentPose():
     """
